function BS = blksetadd(BS, img, step, msk)

% blksetadd -- Add blocks in an image to a blkset
%
% Usage:
%       BS = blksetadd(BS, img, step, msk)
%
% Input:
%       BS          A block set structure to be extended by this call
%       img         Image from which to extract blocks
%       step        A grid structure (see function blockgrids) or a
%                   two-vector containing row step and col step values
%       msk         A mask image indicating positions of pixels to
%                   be excluded (i.e. any blocks containing these
%                   pixels) from the block set
%
% Output:
%       BS          A struct containing information on the block set

Nbr = BS.Bd(1);
Nbc = BS.Bd(2);
[Nir Nic Nih] = size(img);
if nargin < 4,
  msk = [];
end
if nargin < 3,
  step = [1 1];
end

if isstruct(step),
  % Add up number of blocks in grid set
  GRD = step;
  Ng = size(GRD.Grids,1);
  Nib = 0;
  for g = 1:Ng,
    Nib = Nib + prod(size(GRD.Grids{g,1}));
  end
else
  % Compute number of blocks in image
  Nsr = step(1);
  Nsc = step(2);
  Nib = length(1:Nsr:(Nir-Nbr+1))*length(1:Nsc:(Nic-Nbc+1));
end

% Compute number of blocks available in currently allocated space
Nrb = size(BS.BlkMx,2) - BS.bn;
% If image contains more blocks than available, allocate more space
if Nib > Nrb,
  BlkMx = zeros(Nbr*Nbc*Nih, max(2*size(BS.BlkMx,2), Nib - Nrb));
  if ~isempty(BS.BlkMx),
    BlkMx(:,1:BS.bn) = BS.BlkMx(:,1:BS.bn);
  end
  BS.BlkMx = BlkMx;
  Len = zeros(1,max(2*size(BS.Len,1), Nib - Nrb));
  if ~isempty(BS.Len),
    Len(:,1:BS.bn) = BS.Len(:,1:BS.bn);
  end
  BS.Len = Len;
end

if isstruct(step),
  
  for g = 1:Ng,
    for l = 1:size(GRD.Grids{g,1},1),
      for m = 1:size(GRD.Grids{g,1},2),
        r1 = GRD.Grids{g,1}(l,m);
        c1 = GRD.Grids{g,2}(l,m);
        r2 = r1 + Nbr - 1;
        c2 = c1 + Nbc - 1;   
        if r1 > 0 &&  r2 <= Nir && c1 > 0 && c2 <= Nic,
          v = img(r1:r2,c1:c2,:);
          if ~isempty(msk) && sum(vec(msk(r1:r2,c1:c2)) ~= 1) > 0,
            continue;
          end
          BS.Len(BS.bn+1) = norm(v(:));
          % Normalize block if requested
          if BS.nflg,
            v = v(:)/norm(v(:));
	  else
	    v = v(:);
          end
          BS.BlkMx(:,BS.bn+1) = v;
          BS.bn = BS.bn + 1;
        end
      end
    end
  end
    
else

  % Loop over all blocks in image, adding each one to block set
  for l=1:Nsr:(Nir-Nbr+1),
    for m=1:Nsc:(Nic-Nbc+1),
      if ~isempty(msk) && sum(vec(msk(l:(l+Nbr-1),m:(m+Nbc-1))) ~= 1) > 0,
        continue;
      end
      v = img(l:(l+Nbr-1),m:(m+Nbc-1),:);
      BS.Len(BS.bn+1) = norm(v(:));
      % Normalize block if requested
      if BS.nflg,
        v = v(:)/norm(v(:));
      else
	v = v(:);
      end
      BS.BlkMx(:,BS.bn+1) = v;
      BS.bn = BS.bn + 1;
    end
  end

end

return
