function GRD = blockgrids(img, Nbr, Nbc, Vs, cf)

% blockgrids -- Construct a set of block grids covering an image
%
% Usage:
%       GRD = blockgrids(img, Nbr, Nbc, Vs, cf)
%
% Input:
%       img          The image on which the grids are required
%       Nbr          Rows in block
%       Nbc          Cols in block
%       Vs           Vector of grid offsets
%       cf           Flag indicating whether to generate extra
%                    grids as necessary for full image coverage
%
% Output:
%       GRD          A structure containing the following fields
%                      Grid   A cell array containing the grids
%                      Nbr    Rows in block
%                      Nbc    Cols in block
%                      Vs     Vector of grid offsets
%                      BlkCnt An image containing the counts of
%                             number of blocks covering each pixel
%
% This function constructs a set of grids covering the specified
% image. The parameters Nbr and Nbc specify the block size of the
% grid, and Vs is a matrix of offsets indicating the offset of each
% grid from the origin. If the requested grids do not cover the
% entire image, additional grids are automatically added as needed
% to give full coverage. The returned cell array is of the form {R1
% C1; R2 C2; ...} where Rn and Cn are matrices of row and column
% coordinates for grid n.

% If cf is not specified, default to generating extra grids to
% cover the image if necessary
if nargin < 5,
  cf = 1;
end

% If Vs not specified, default to requesting a single grid at the origin
if nargin < 4,
  Vs = [1 1];
end

% Transpose Vs if incorrectly specified
if size(Vs,1) < size(Vs,2),
  Vs = Vs';
end

% If necessary, copy row and column offsets to be the same
if size(Vs,2) == 1,
  Vs = [Vs Vs];
end

% Initialize
Ng = size(Vs,1);
Grids = cell(Ng,2);
[Nir Nic Nih] = size(img);
nrm = zeros(Nir,Nic);

% Iterate over all grids
for g = 1:Ng,
  Ngr = Vs(g,1); % Current row grid offset
  Ngc = Vs(g,2); % Current col grid offset

  Nr = floor((Nir+1-Ngr)/Nbr); % Number of blocks in row
  Nc = floor((Nic+1-Ngc)/Nbc); % Number of blocks in col

  RR = Ngr + Nbr*(0:(Nr-1)); % Row indices of block origin
  RC = Ngc + Nbc*(0:(Nc-1)); % Col indices of block origin

  % Construct grid
  [GC, GR] = meshgrid(RC, RR);
  Grids{g,1} = GR;
  Grids{g,2} = GC;

  % Increment block counter for current grid
  r1 = GR(1,1);
  r2 = GR(end,end)+Nbr-1;
  c1 = GC(1,1);
  c2 = GC(end,end)+Nbc-1;
  nrm(r1:r2,c1:c2) = nrm(r1:r2,c1:c2) + 1;
end

if cf,
  % Construct additional grids for uncovered pixels at image end
  if sum(nrm(:,end) == 0) > sum(nrm(end,:) == 0),
    [Grids, nrm] = endcol(Grids, nrm, Nbr, Nbc);
    [Grids, nrm] = endrow(Grids, nrm, Nbr, Nbc);
  else
    [Grids, nrm] = endrow(Grids, nrm, Nbr, Nbc);
    [Grids, nrm] = endcol(Grids, nrm, Nbr, Nbc);
  end

  % Construct final grid containing a single block at bottom right of
  % image if any pixels remain uncovered by any block
  if nrm(end,end) == 0,
    g = size(Grids,1) + 1;
    Grids{g,1} = Nir-Nbr;
    Grids{g,2} = Nic-Nbc;
    nrm(Grids{g,1}:end,Grids{g,2}:end) = nrm(Grids{g,1}:end,Grids{g,2}:end)+1;
  end
end


GRD = struct('Grids', {Grids}, 'Nbr', Nbr, 'Nbc', Nbc, ...
             'Vs', Vs, 'BlkCnt', nrm);

return


function [Grids, nrm] = endcol(Grids, nrm, Nbr, Nbc)

  % Construct additional grid for uncovered image column at row end
  if sum(nrm(:,end) == 0) > 0,
    RR = [];
    RC = size(nrm,2)-Nbc+1;
    % Construct column of blocks, omitting those that are not required
    for r = 1:Nbr:(size(nrm,1)-Nbr+1),
      if sum(nrm(r:(r+Nbr-1),end) == 0) > 0,
	RR = [RR r];
	nrm(r:(r+Nbr-1),RC:end) = nrm(r:(r+Nbr-1),RC:end) + 1;
      end
    end
    if ~isempty(RR),
      % Insert last block in image if required and does not overlap
      % with existing blocks in this grid
      if (RR(end)+Nbr-1 < size(nrm,1)-Nbr+1) && ...
	    sum(nrm((end-Nbr+1):end,end) == 0) > 0
	RR = [RR size(nrm,1)-Nbr+1];
	nrm((end-Nbr+1):end,(end-Nbc+1):end) = ...
	    nrm((end-Nbr+1):end,(end-Nbc+1):end) + 1;
      end
      [GC, GR] = meshgrid(RC, RR);
      g = size(Grids,1) + 1;
      Grids{g,1} = GR;
      Grids{g,2} = GC;
    end
  end

return


function [Grids, nrm] = endrow(Grids, nrm, Nbr, Nbc)

  % Construct additional grid for uncovered image row at col end
  if sum(nrm(end,:) == 0) > 0,
    RR = size(nrm,1)-Nbr+1;
    RC = [];
    % Construct row of blocks, omitting those that are not required
    for c = 1:Nbc:(size(nrm,2)-Nbc+1),
      if sum(nrm(end,c:(c+Nbc-1)) == 0) > 0,
	RC = [RC c];
	nrm(RR:end,c:(c+Nbc-1)) = nrm(RR:end,c:(c+Nbc-1)) + 1;
      end
    end
    if ~isempty(RC),
      % Insert last block in image if required and does not overlap
      % with existing blocks in this grid
      if (RC(end)+Nbc-1 < size(nrm,2)-Nbc+1) && ...
	    sum(nrm(end,(end-Nbc+1):end) == 0) > 0
	RC = [RC size(nrm,2)-Nbc+1];
	nrm((end-Nbr+1):end,(end-Nbc+1):end) = ...
	    nrm((end-Nbr+1):end,(end-Nbc+1):end) + 1;
      end
      [GC, GR] = meshgrid(RC, RR);
      g = size(Grids,1) + 1;
      Grids{g,1} = GR;
      Grids{g,2} = GC;
    end
  end

return
