function IB = imageblocks(img, Bs, fn)
  
% imageblocks -- Construct a struct representing all blocks (of
%                specified size) in the image
%
% Usage:
%       IB = imageblocks(img, Bs, fn)
%
% Input:
%       img         Image from which to extract blocks
%       Bs          Block size vector
%       fn          Optional function operating on block before extraction
%
% Output:
%       IB           A struct containing information on the block set


if nargin < 3,
  fn = [];
end

Nbr = Bs(1);
Nbc = Bs(2);
[Nir Nic Nih] = size(img);
Nib = length(1:(Nir-Nbr+1))*length(1:(Nic-Nbc+1));   
if isempty(fn),
  Nv = Nbr*Nbc*Nih;
else
  Nv = prod(size(fn(img(1:Nbr,1:Nbc,1:Nih))));
end

BlkMx = zeros(Nv, Nib);
BlkCrd = zeros(2, Nib);

% Loop over all blocks in image, adding each one to block set
n = 1;
for l=1:(Nir-Nbr+1),
  for m=1:(Nic-Nbc+1),
    v = img(l:(l+Nbr-1),m:(m+Nbc-1),1:Nih);
    if isempty(fn),
      BlkMx(:,n) = v(:);
    else
      BlkMx(:,n) = vec(fn(v));
    end
    BlkCrd(:,n) = [l m];
    n = n + 1;
  end
end

IB = struct('Nbr', Nbr, 'Nbc', Nbc, 'BlkMx', BlkMx, 'BlkCrd', BlkCrd,...
            'Fn', fn);

return
