function [M, U12] = blkmatch(U, T, W, aflg)
  
% blkmatch -- Find matching blocks in an image
%
% Usage:
%       M = blkmatch(U, T, W, aflg)
%
% Input:
%       U           The image in which to find matches
%       T           The block for which matches are required
%       W           A mask block with 0 and 1 entries
%       aflg        A flag selecting the measure of match. The
%                   default, 1, gives an angular measure (the
%                   normalised inner product), while 0 gives the
%                   Euclidean distance. Note that the angular measure
%                   is not a distance, and also that the caching of
%                   FFT results will give incorrect values if aflg is
%                   changed between calls to blkmatch using the cached
%                   values.
% 
% Output:
%       M           A map providing the measure of match, at each
%                   position of the image, between the image block
%                   at that position and the block T
%       U12         Cell array containing precomputed FFTs of U and
%                   U.^2

if nargin < 4,
  aflg = 1;
end
if nargin < 3 || isempty(W),
  W = ones(size(T));
end

Tw =  W.*T;
if aflg,
  Twn = Tw / norm(Tw(:));
  Th = flipdim(flipdim(Twn,2),1);
else
  Th = flipdim(flipdim(Tw,2),1);  
end

if iscell(U),
  if U{5} ~= aflg,
    warning(['blkmatch: cached FFT results used, and match measure ' ...
             'has changed between calls']);
  end
  su = U{3};
  ru = U{4};
  U1 = {U{1}, su, ru};
  U2 = {U{2}, su, ru};
else
  if aflg,
    U1 = U / max(U(:));
  else
    U1 = U;
  end
  U2 = U1.^2;
  su = size(U);
  ru = isreal(U);
end
[ip, U1f] = fftconv2(U1, Th, 'same');
ip = sum(ip,3);
[nm, U2f] = fftconv2(U2, flipdim(flipdim(W,2),1), 'same');
nm = sum(abs(nm),3);

if aflg,
  Minf = (nm == 0);
  nm(Minf) = 1;
  ip(Minf) = 0;
  M = ip ./ sqrt(nm);
else
  M = real(sqrt(nm - 2*ip + sum(Tw(:).^2)));
end
  
ofs = floor((size(T)-1)/2)+1;
vbs = su - size(T);
M = M(ofs(1):(ofs(1)+vbs(1)),ofs(2):(ofs(2)+vbs(2)));

if nargout > 1,
  U12 = {U1f, U2f, su, ru, aflg};
end

return
