function BS = blksrcadd(BS, img, msk, Wb)

% blksrcadd -- Add an image to the the fast block search struct
%
% Usage:
%       BS = blksrcadd(BS, img, msk, Wb)
%
% Input:
%       BS          Main struct. If empty, a new one is created.
%       img         Image containing blocks
%       msk         Mask indicating usable region of image
%       Wb          Base block active region mask (should only be
%                   specified when BS is empty)
%
% Output:
%       BS           A struct containing information on the block source

if isempty(BS),
  if nargin < 4,
    Wb = ones(8,8);
  end
  BS = struct('Wb',Wb,'in',0,'ImgC',{{}},'MskC',{{}},'VldC',{{}},...
              'U12C',{{}});
else
  if nargin == 4,
    warning('blksrcadd: Wb should only be specified when BS is empty');
  else
    Wb = BS.Wb;
  end
end
  
if nargin < 3 || isempty(msk),
    msk = ones(size(img,1),size(img,2));
end

if ~isempty(img),
  % Record image and mask
  BS.in = BS.in + 1;
  BS.ImgC{BS.in} = img;
  BS.MskC{BS.in} = msk;
  % Compute indicator, VldC, of positions of valid results in the
  % match map. The match map (where the value at each pixel gives the
  % degree of match of the block with its top left pixel at that
  % position) size depends on the block size, since blocks with
  % pixels overlapping the image boundary are excluded. Results are
  % valid when all pixels in a block are valid according to the
  % image mask
  Vpn = fftconv2(msk, flipdim(flipdim(BS.Wb,2),1), 'same');
  ofs = floor((size(BS.Wb)-1)/2)+1;
  vbs = size(msk) - size(BS.Wb);
  Vpn = Vpn(ofs(1):(ofs(1)+vbs(1)),ofs(2):(ofs(2)+vbs(2)));
  BS.VldC{BS.in} = (round(Vpn) == sum(Wb(:)));
  % Record coordinate set for match map for this image
  [Gc,Gr] = meshgrid(1:size(Vpn,2),1:size(Vpn,1));
  BS.Gr{BS.in} = Gr;
  BS.Gc{BS.in} = Gc;
  % Initialise cache for DFT of image and square of image
  BS.U12C{BS.in} = {};
end

return
