function [U, D, Idx] = nneuclid(V, u, m, d, w)

% nneuclid -- Find a set of neighbouring vectors according to
%             Euclidean distance
%
% Usage:
%       [U, D, Idx] = nneuclid(V, u, m, d, w)
%
% Input:
%       V           Vectors to search arranged as rows of a matrix
%       u           The reference vector for which neighbours are desired
%       m           Maximum number of neighbours
%       d           Neighbourhood radius (Euclidean distance)
%       w           Weighting vector
%
% Output:
%       U           A matrix containing the neighbour vectors
%       D           A vector giving distances to the neighbour vectors
%       Idx         A vector giving indices in block set of found vectors

[Nd,Nv] = size(V);  

if nargin < 5 || isempty(w),
  w = ones(size(u));
end
if nargin < 4,
  d = [];
end
if nargin < 3,
  m = [];
end  

if size(u,1) < size(u,2),
  u = u';
end
if size(w,1) < size(w,2),
  w = w';
end
if m > Nv,
  m = Nv;
end

Da = sqrt(sum((V'*diag(w)).^2,2) - 2*V'*(w.*u) + sum((w.*u).^2));
[D, Idx] = sort(Da);

if ~isempty(m),
  Idx = Idx(1:m);
  D = D(1:m);
end

if ~isempty(d),
  Idx = Idx(D < d);
  D = D(D < d);
end

U = V(:,Idx);

return
