% jntoptinpaint -- Perform inpainting via the method described in
%                  B. Wohlberg, "Inpainting by Joint Optimization
%                  of Linear Combinations of Exemplars", IEEE
%                  Signal Processing Letters
%
% Usage:
%       [ipt, iptinf] = jntoptinpaint(img, msk, opt)
%
% Input:
%       img         Image to be inpainted
%       msk         Inpainting region mask
%       opt         Structure containing algorithm parameters
%
% Output:
%       ipt         Inpainted image
%       iptinf      Structure containing details of inpainting process

function [ipt, iptinf] = jntoptinpaint(img, msk, opt)

% Start timer
tstart = tic;

% Set parameters
if nargin < 3,
  opt = [];
end
opt = defaultopts(opt);
icgt = opt.InitCGTol;
lambda = opt.BPDNLambda;
vrb = opt.Verbose;

% Construct inpainting structure
if isfield(opt,'INPT'),
  INPT = opt.INPT; 
else
  if vrb,
    disp('Constructing dictionaries');
  end
  INPT = initinpnt(img, msk, opt);
  INPT.gamma0 = opt.FuncGamma(1);
  INPT.gamma1 = opt.FuncGamma(2);
  INPT.imgrng = [min(img(:)) max(img(:))];
  INPT.ClipGrids = opt.ClipGrids;
end

% Construct initial solution via CG
if vrb,
  disp(sprintf('Constructing initial CG solution  (problem size %dx%d)',...
		 [INPT.Nv1+INPT.Nv2, INPT.Nv0]));
end
Ng = size(INPT.Phi,1);
b = [sqrt(INPT.gamma0/INPT.Nv1)*inpntRB(img, INPT); zeros(INPT.Nv2,1)];
ATb = inpntAT(b, INPT);
ATA = @(x) inpntAT(inpntA(x, INPT), INPT);
[alpha0, cgf] = cgs(ATA, ATb, icgt, 20000);

% Solve using single IRLS BPDN iterations followed by dictionary update 
n = opt.CoefShrinkIter;
m = opt.BPDNOpt.MaxMainIter - n;
opt.BPDNOpt.MaxMainIter = 1;
optinf = [];
while n > 0,
  A = @(x) inpntA(x, INPT);
  AT = @(x) inpntAT(x, INPT);
  if vrb,
    disp(sprintf('Performing sparse optimisation (problem size %dx%d)',...
		 [INPT.Nv1+INPT.Nv2, INPT.Nv0]));
  end
  [alpha, optinf] = irlsbpdn({A,AT}, b, lambda, opt.BPDNOpt, alpha0);
  [INPT, alpha0] = updateinpnt(INPT, alpha, opt);
  %opt.BPDNOpt = optinf.opt;
  n = n-1;
end
opt.BPDNOpt.MaxMainIter = m;


% Solve using IRLS BPDN
if opt.BPDNOpt.MaxMainIter > 0,
  A = @(x) inpntA(x, INPT);
  AT = @(x) inpntAT(x, INPT);
  if vrb,
    disp(sprintf('Performing sparse optimisation (problem size %dx%d)',...
		 [INPT.Nv1+INPT.Nv2, INPT.Nv0]));
  end
  [alpha, optinf] = irlsbpdn({A,AT}, b, lambda, opt.BPDNOpt, alpha0);
else
  alpha = alpha0;
end


% Reconstruct image
[rgnipt, rgnbnm, rgnbvr, rgngrd, grdalp] = inpntimg(alpha, INPT);
ipt = img;
mskh = repmat(INPT.iptmsk, [1 1 size(img,3)]);
ipt(~mskh) = rgnipt(~mskh);

iptinf.opt = opt;
iptinf.optinf = optinf;
iptinf.ipt = INPT;
iptinf.alpha = alpha;
iptinf.balpha = grdalp;
iptinf.dtfidv = inpntA(alpha, INPT);
iptinf.b = b;
iptinf.rgnbnm = rgnbnm;
iptinf.rgnbvr = rgnbvr;
iptinf.rgngrd = rgngrd;
iptinf.runtime = toc(tstart);

return


function opt = defaultopts(opt)
  
  if ~isfield(opt,'Verbose'),
    opt.Verbose = 0;
  end
  if ~isfield(opt,'BlockSize'),
    opt.BlockSize = [8 8];
  end
  if ~isfield(opt,'GridStep'),
    opt.GridStep = [1 floor(min(opt.BlockSize)/2)];
  end
  if ~isfield(opt,'FuncGamma'),
    opt.FuncGamma = [1.0 1.0];
  end
  if ~isfield(opt,'BlkSrcRotations'),
    opt.BlkSrcRotations = [0];
  end
  if ~isfield(opt,'MeanSubBlocks'),
    opt.MeanSubBlocks = 0;
  end
  if ~isfield(opt,'NormaliseBlocks'),
    opt.NormaliseBlocks = 0;
  end
  if ~isfield(opt,'ClipGrids'),
    opt.ClipGrids = 0;
  end
  if ~isfield(opt,'ExtraGrids'),
    opt.ExtraGrids = 0;
  end
  if ~isfield(opt,'BlockSourceMask'),
    opt.BlockSourceMask = [];
  end
  if ~isfield(opt,'PartialInpaint'),
    opt.PartialInpaint = 0;
  end
  if ~isfield(opt,'UseRegionEstimate'),
    opt.UseRegionEstimate = 0;
  end
  if ~isfield(opt,'NNMinKnownFrac'),
    opt.NNMinKnownFrac = 0.75;
  end
  if ~isfield(opt,'NNDistMeasure'),
    opt.NNDistMeasure = 'absang';
  end
  if ~isfield(opt,'NNMaxDist'),
    opt.NNMaxDist = 0.3;
  end
  if ~isfield(opt,'NNDictSize'),
    opt.NNDictSize = [];
  end
  if ~isfield(opt,'NNMaxDictSize'),
    opt.NNMaxDictSize = 8*prod(opt.BlockSize);
  end
  if ~isfield(opt,'IKPBMaxKnownFrac'),
    opt.IKPBMaxKnownFrac = opt.NNMinKnownFrac;
  end
  if ~isfield(opt,'IKPBMaxDictSize'),
    opt.IKPBMaxDictSize = 4*prod(opt.BlockSize);
  end
  if ~isfield(opt,'IKPBSearchRadius'),
    opt.IKPBSearchRadius = 20;
  end
  if ~isfield(opt,'UseKSVDDict'),
    opt.UseKSVDDict = 0;
  end
  if ~isfield(opt,'KSVDDictSize'),
    opt.KSVDDictSize = 8*prod(opt.BlockSize);
  end
  if ~isfield(opt,'MaxDictSize'),
    opt.MaxDictSize = [];
  end
  if ~isfield(opt,'DictPruneTol'),
    opt.DictPruneTol = 1e-7;
  end
  if ~isfield(opt,'CoefShrinkFactor'),
    opt.CoefShrinkFactor = 1e-2;
  end
  if ~isfield(opt,'CoefShrinkIter'),
    opt.CoefShrinkIter = 0;
  end
  if ~isfield(opt,'InitCGTol'),
    opt.InitCGTol = 1e-3;
  end
  if ~isfield(opt,'BPDNLambda'),
    opt.BPDNLambda = 1e-5;
  end
  if ~isfield(opt,'BPDNOpt'),
    opt.BPDNOpt = [];
  end 
  if ~isfield(opt.BPDNOpt,'Verbose'),
   opt.BPDNOpt.Verbose = opt.Verbose;
  end
  if ~isfield(opt.BPDNOpt,'MaxMainIter'),
   opt.BPDNOpt.MaxMainIter = 5;
  end

return


function INPT = initinpnt(img, msk, opt)

% initinpnt -- Initialise a structure containting the information
%              necessary for exemplar-based inpainting
%
% Usage:
%       INPT = initinpnt(img, msk, opt)
%
% Input:
%       img         Image to be inpainted
%       msk         Region mask image
%       opt         Options structure
%
% Output:
%       INPT        A struct representing the inpainting information

Nbr = opt.BlockSize(1);
Nbc = opt.BlockSize(2);
Nbb = size(img,3);

% Set up block source and inpainting region masks
if isempty(opt.BlockSourceMask),
  srcmsk = msk;
else
  srcmsk = opt.BlockSourceMask;
end
iptmsk = msk; % This is actually the complement of a mask
iptmskh = repband(iptmsk,Nbb);

% Construct grid sets
GRD = blockgrids(img, Nbr, Nbc, opt.GridStep);
[GRDi, GRDk] = blockgridsmask(GRD, iptmsk);


if opt.ExtraGrids,
  
  % User selected grid offsets
  Vs = GRDi.Vs;
  % Construct list of unselected offsets
  Vm = ones(Nbr, Nbc);
  Vm(Vs(:,1) + Nbr*(Vs(:,2)-1)) = 0;
  [Vr, Vc] = ndgrid(1:Nbr, 1:Nbc);
  Ve = [Vr(:) Vc(:)];
  Ve = Ve(Vm == 1,:);
  % Construct grids corresponding to unselected offsets
  GRDe = blockgrids(img, Nbr, Nbc, Ve, 0);

  GRDe = blockgridsmask(GRDe, iptmsk);

  egv = [];
  for g = 1:size(GRDe.Grids,1),
    for l = 1:size(GRDe.Grids{g,1},1),
      for m = 1:size(GRDe.Grids{g,1},2),
        % Extract top-left coordinates of current block
        r1 = GRDe.Grids{g,1}(l,m);
        c1 = GRDe.Grids{g,2}(l,m);
        % Process current block if not marked as removed from grid
        if r1 ~= 0 && c1 ~= 0,
          % Construct bottom-right coordinates of current block
          r2 = r1 + Nbr - 1;
          c2 = c1 + Nbc - 1;
          mskglm = iptmsk(r1:r2,c1:c2); % Current block known region mask
          % Compute fraction of known pixels in current block
          kpf = (sum(vec(mskglm) == 1)/prod(size(mskglm)));
	  % If block known pixels fraction is less then threshold,
          % mark block as deleted from grid and decrement the
          % number blocks covering each position counter for all
          % pixels in that block
          if kpf < opt.NNMinKnownFrac,
            GRDe.Grids{g,1}(l,m) = 0;
            GRDe.Grids{g,2}(l,m) = 0;
            GRDe.BlkCnt(r1:r2,c1:c2) = GRDe.BlkCnt(r1:r2,c1:c2) - 1;
          end
        end
      end
    end
    % Record indices of grids that have no remaining blocks
    if sum(vec(GRDe.Grids{g,1})) == 0,
      egv = [egv g];
    end
  end
 
  % Delete completely empty grids
  GRDe.Grids(egv,:) = [];

  GRDe.Vs(egv,:) = [];
  
  % Append extra grids onto main grid set
  GRDi.Grids = cat(1, GRDi.Grids, GRDe.Grids);
  GRDi.Vs = cat(1, GRDi.Vs, GRDe.Vs);
  GRDi.BlkCnt = GRDi.BlkCnt + GRDe.BlkCnt;
end


%disp(sprintf('  Using %d grids of %dx%d blocks', size(GRDi.Grids,1), ...
%	     Nbr, Nbc));
%for g = 1:size(GRDi.Grids,1),
%  disp(sprintf('    Grid %d (offset %d,%d) has %d blocks', ...
%	       g, GRDi.Vs(g,1), GRDi.Vs(g,2), sum(vec(GRDi.Grids{g,1}~= 0))));
%end
  


% If PartialInpaint enabled, only blocks with at least
% opt.NNMinKnownFrac are inpainted, the other blocks being deleted
% from the inpainting grids. 
egv = [];
if opt.PartialInpaint,
  for g = 1:size(GRDi.Grids,1),
    for l = 1:size(GRDi.Grids{g,1},1),
      for m = 1:size(GRDi.Grids{g,1},2),
        % Extract top-left coordinates of current block
        r1 = GRDi.Grids{g,1}(l,m);
        c1 = GRDi.Grids{g,2}(l,m);
        % Process current block if not marked as removed from grid
        if r1 ~= 0 && c1 ~= 0,
          % Construct bottom-right coordinates of current block
          r2 = r1 + Nbr - 1;
          c2 = c1 + Nbc - 1;
          mskglm = iptmsk(r1:r2,c1:c2); % Current block known region mask
          % Compute fraction of known pixels in current block
          kpf = (sum(vec(mskglm) == 1)/prod(size(mskglm)));
	  % If block known pixels fraction is less then threshold,
          % mark block as deleted from grid and decrement the
          % number blocks covering each position counter for all
          % pixels in that block
          if kpf < opt.NNMinKnownFrac,
            GRDi.Grids{g,1}(l,m) = 0;
            GRDi.Grids{g,2}(l,m) = 0;
            GRDi.BlkCnt(r1:r2,c1:c2) = GRDi.BlkCnt(r1:r2,c1:c2) - 1;
          end
        end
      end
    end
    % Record indices of grids that have no remaining blocks
    if sum(vec(GRDi.Grids{g,1})) == 0,
      egv = [egv g];
    end
  end
  % In the inpainting mask region, delete any part that has no
  % inpainting blocks
  iptmsk(GRDi.BlkCnt == 0) = 1;
end
% Delete completely empty grids
GRDi.Grids(egv,:) = [];


% Initialise variables data structures
Ng = size(GRDi.Grids,1); % Number of grids
CRgn = cell(Ng,1);       % Known region mask for each block
CPhi = cell(Ng,1);       % Dictionary for each block
CPSz = cell(Ng,1);       % Dictionary component sizes for each block
CCrd = cell(Ng,1);       % Image coordinates of each block
CGms = cell(Ng,1);       % Inpainting region mask for each grid
CIx0 = cell(Ng,1);       % Coordinates for each block in
                         % concatenated coefficient vector
CIx1 = cell(Ng,1);       % Coordinates for each block in
                         % concatenated known-region vector
CMap = cell(Ng,1);       % Map from inpainting region vector for each block
CGrm = cell(Ng,1);       % Inpainting region vector mask for each grid

% Initialise vector coordinate counters
i0 = 1;
i1 = 1;

BSk = [];
if opt.UseKSVDDict,
  BSk = blksetinit([Nbr Nbc], 1);
else
  [icy,icx] = ndgrid(1:size(img,1),1:size(img,2));
  ICC = {};
end
BS = blksrcadd([], [], [], ones(Nbr,Nbc));
for ra = opt.BlkSrcRotations,
  if ra == 0,
    BS = blksrcadd(BS, img, srcmsk == 1);
    if opt.UseKSVDDict,
      BSk = blksetadd(BSk, img, [1 1], srcmsk);
    else
      ICC = cat(1,ICC, {icy, icx});
    end
  else
    imgr = imrotate(img, ra, 'bilinear');
    mskr = imrotate(srcmsk, ra, 'bilinear');
    BS = blksrcadd(BS, imgr, mskr > 0.99);
    if opt.UseKSVDDict,
      BSk = blksetadd(BSk, imgr, [1 1], mskr);
    else
      icyr = imrotate(icy, ra, 'bilinear');
      icxr = imrotate(icx, ra, 'bilinear');
      ICC = cat(1,ICC, {icyr, icxr});
    end
  end
end

if opt.UseKSVDDict,
  BSk = blksetend(BSk);
  if opt.Verbose,
    disp('Constructing KSVD dictionary');
  end
  D = ksvddict(img, BSk, opt.KSVDDictSize);
  BSk = [];
end

Ntb = 0;

% Iterate over all grids
for g = 1:Ng,

  % Initialise per-grid data structures
  Nb = sum(sum(GRDi.Grids{g,1} > 0));
  CRgn{g} = cell(Nb,1);
  CPhi{g} = cell(Nb,1);
  CPSz{g} = zeros(Nb,3);
  CCrd{g} = zeros(Nb,4);
  CGms{g} = zeros(size(iptmskh));
  CIxB{g} = zeros(Nb,2);
  CIxC{g} = zeros(Nb,2);
  CMap{g} = cell(Nb,1);

  k = 1; % Initialise block counter for current grid

  % Iterate over all blocks in current grid
  for l = 1:size(GRDi.Grids{g,1},1),
    for m = 1:size(GRDi.Grids{g,1},2),

      % Extract top-left coordinates of current block
      r1 = GRDi.Grids{g,1}(l,m);
      c1 = GRDi.Grids{g,2}(l,m);

      % Process current block if not marked as removed from grid
      if r1 ~= 0 && c1 ~= 0,

	% Construct bottom-right coordinates of current block
	r2 = r1 + Nbr - 1;
	c2 = c1 + Nbc - 1;
        CRgn{g}{k} = iptmskh(r1:r2,c1:c2,:); % Current block known region mask

        % Compute fraction of known pixels in current block
        kpf = (sum(vec(CRgn{g}{k}) == 1)/prod(size(CRgn{g}{k})));
        
        % Construct nearest neighbour dictionary
	Bnn = [];
        if kpf >= opt.NNMinKnownFrac,
          % If sufficient fraction of pixels in block are known,
          % construct dictionary via nearest neighbour search in
          % provided block set
	  u = img(r1:r2,c1:c2,:);
	  w = iptmsk(r1:r2,c1:c2);
          % Determine nearest neighbour dictionary size
          if isempty(opt.NNDictSize),
            Nds = opt.NNMaxDictSize;
          else
            Nds = floor(opt.NNDictSize(1) + opt.NNDictSize(2)*(1-kpf) + ...
                        opt.NNDictSize(3)*(1-kpf).^2);
          end
          if ~isempty(opt.MaxDictSize) && Nds > opt.MaxDictSize,
            Nds = opt.MaxDictSize;
          end
          
          % Construct nearest neighbour dictionary
          Bnn = blksrcmatch(BS,u,Nds,opt.NNMaxDist,w,opt.NNDistMeasure);
        end

        % Construct spatial neighbourhood dictionary
        Bsn = [];

	if opt.UseRegionEstimate, 

	   u = img(r1:r2,c1:c2,:);
	   Bsn = blksrcmatch(BS, u, opt.IKPBMaxDictSize, ...
			     opt.NNMaxDist, [], opt.NNDistMeasure);

        elseif kpf < opt.IKPBMaxKnownFrac || size(Bnn,2) == 0,
          % If insufficient fraction of pixels in block are known,
          % construct a dictionary by scanning the spatial
          % neighbourhood of the current block in a spiral path,
          % adding blocks without any unknown pixels
	  if opt.UseKSVDDict,
	    B = D;
	  else
	    Bsn = spatialnbhddict(BS,ICC,r1,c1,r2,c2,opt.IKPBSearchRadius,...
                                  opt.IKPBMaxDictSize);
	  end
        end
   
        % Construct joint dictionary
        if isempty(opt.MaxDictSize),
          Nsn = size(Bsn,2);
        else
          Nsn = opt.MaxDictSize - size(Bnn,2);
        end
        if size(Bsn,2) == 0,
          Nsn = 0;
        end
        if size(Bsn,2) < Nsn,
          Nsn = size(Bsn,2);
        end

        B = [Bnn Bsn(:,1:Nsn)];
        CPSz{g}(k,1:2) = [size(Bnn,2) Nsn];

	if opt.MeanSubBlocks,
	  B = B - repmat(mean(B),size(B,1),1);
	end
	if opt.NormaliseBlocks,
          B = normdict(B);
        end
	if opt.MeanSubBlocks,
	  dc = 1e1*(norm(B(:,1))/sqrt(size(B,1)))*ones(size(B,1),1);
	  B = [dc B];
	end
	CPhi{g}{k} = prunedict(B, opt.DictPruneTol);
        CPSz{g}(k,3) = size(CPhi{g}{k},2);
        
	CCrd{g}(k,:) = [r1 c1 r2 c2]; % Current block coordinates
	CGms{g}(r1:r2,c1:c2,:) = 1; % Mark current block in grid mask
 
	% Construct inpainting vector to block map for current block
        bx = zeros(size(img));
        bx(r1:r2,c1:c2,:) = reshape(1:((r2-r1+1)*(c2-c1+1)*Nbb),...
				    r2-r1+1,c2-c1+1,Nbb);
        CMap{g}{k} = [bx((bx>0).*(1-iptmskh) > 0) find(bx(iptmskh == 0))];

	% Compute dictionary coefficient vector size for current block
        n0 = size(CPhi{g}{k},2);
	% Compute known region vector size for current block
        n1 = sum(CRgn{g}{k}(:) == 1);

	% Construct vector coordinate records
	CIx0{g}(k,:) = [i0 i0+n0-1];
	CIx1{g}(k,:) = [i1 i1+n1-1];

	% Increment vector coordinate counters
	i0 = i0 + n0;
	i1 = i1 + n1;

	% Increment current block counter
	k = k+1;

      end

    end
  end

  Ntb = Ntb + k-1;

  % Compute inpainting region vector mask for current grid
  CGrm{g} = CGms{g}(iptmskh == 0);

end

Nv0 = CIx0{Ng}(end,2); % Concatenated dict coefficient vector length
Nv1 = CIx1{Ng}(end,2); % Concatenated known region vector length
Nvg = size(CGrm{g},1); % Inpainting region vector length
Nv2 = (Ng-1)*Nvg;      % Overlap constraint vector length

if opt.Verbose,
  
  disp(sprintf('  Using %d grids of %dx%d blocks', Ng, Nbr, Nbc));
  for g = 1:Ng,
    disp(sprintf('    Grid %d (offset %d,%d) has %d blocks (%d have NND)', ...
		 g, GRDi.Vs(g,1), GRDi.Vs(g,2), size(CRgn{g},1), ...
	         sum(CPSz{g}(:,1) > 0)));
  end
  
  disp('  Vector lengths');
  disp(sprintf('    Concatentated dict. coeff.: %6d', Nv0));
  disp(sprintf('    Known region penalty      : %6d', Nv1));
  disp(sprintf('    Block overlap penalty     : %6d', Nv2));
  
end


% Construct output structure
INPT = struct('Rgn', {CRgn}, 'Phi', {CPhi}, 'Crd', {CCrd}, 'Gms', {CGms}, ...
              'Ix0', {CIx0}, 'Nv0', Nv0, 'Ix1', {CIx1}, 'Nv1', Nv1, ...
	      'Nv2', Nv2, 'Map', {CMap}, 'Grm', {CGrm}, 'Nvg', Nvg, ...
	      'CPSz', {CPSz}, 'Nih', size(img,3), 'Ntb', Ntb, ...
	      'BS', BS, 'iptmsk', iptmsk);

return



function B = spatialnbhddict(BS, ICC, r1, c1, r2, c2, scr, mds)
    
  ofs = floor((size(BS.Wb)-1)/2)+1;
  Wbx = repmat(BS.Wb,[1 1 size(BS.ImgC{1},3)]);
  Sd = cell(BS.in,1);
  for k = 1:BS.in,
    icy = ICC{k,1};
    icx = ICC{k,2};
    snm = sqrt((icy-(r1+r2)/2).^2 + (icx-(c1+c2)/2).^2);
    vbs = size(BS.MskC{k}) - size(BS.Wb);
    Sd{k} = snm(ofs(1):(ofs(1)+vbs(1)),ofs(2):(ofs(2)+vbs(2)));
    Sd{k}(BS.VldC{k}~=1) = Inf;
  end

  B = [];
  r = 1;
  while r <= scr && size(B,2) <= mds,
    for k = 1:BS.in,
      Mc{k} = (1-double(Sd{k} > (r-1) & Sd{k} <= r));
    end
    Br = mapblkdict(BS.ImgC, Wbx, Mc, BS.Gr, BS.Gc, [], 0.5);
    B = [B Br];
    r = r + 1;
  end

return


function D = ksvddict(img, BS, dsz)
  params.dictsize = dsz;
  params.data = BS.BlkMx;
  params.iternum = 50;
  params.memusage = 'high';
  params.Tdata = 6;
  params.codemode = 'sparsity';
  [D,g,err] = ksvd(params,'');
return


function [INPT, alpha] = updateinpnt(INPT, alpha, opt)

Ng = size(INPT.Phi,1);
au = cell(Ng,1);
Bu = cell(Ng,1);
cu = cell(Ng,1);
i0 = 1;
% Iterate over all grids
for g = 1:Ng,
  Nb = size(INPT.Crd{g},1); % Number of blocks in current grid
  au{g} = cell(Nb,1);
  Bu{g} = cell(Nb,1);
  % Iterate over all blocks in current grid
  for k = 1:Nb,

    av = alpha(inpntcx(g, k, INPT));
    avs = abs(av)>max(abs(av))*opt.CoefShrinkFactor;
    au{g}{k} = av(avs);
    Bu{g}{k} = INPT.Phi{g}{k}(:,avs);

    % New dictionary coefficient vector size for current block
    n0 = size(Bu{g}{k},2);
    % New vector coordinate record
    cu{g}(k,:) = [i0 i0+n0-1];
    % Increment vector coordinate counter
    i0 = i0 + n0;

  end
end

INPT.Ix0 = cu;
INPT.Nv0 = cu{Ng}(end,2);
alpha = zeros(INPT.Nv0,1);
for g = 1:Ng,
  Nb = size(INPT.Crd{g},1);
  for k = 1:Nb,
    alpha(inpntcx(g, k, INPT)) = au{g}{k};
    INPT.Phi{g}{k} = Bu{g}{k};
  end
end

return


function u = inpntA(alpha, INPT)

% inpntA -- Apply inpainting operator A to compound coefficient vector
%
% Usage:
%       u = inpntA(alpha, INPT)
%
% Input:
%       alpha        Compound coefficient vector
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       u            Compound vector resulting from operator A

Ng = size(INPT.Phi,1); % Number of grids
u1 = zeros(INPT.Nv1,1); % Known pixel mismatch vector
u2 = zeros((Ng-1)*INPT.Nvg,1); % Grid overlap mismatch vector

% Initialise cell array of sum of QT applied to blk recons for each grid
QTUB = cell(Ng,1);

% Iterate over all grids
for g = 1:Ng,
  Nb = size(INPT.Crd{g},1); % Number of blocks in current grid
  qtub = zeros(INPT.Nvg,1); % Initialise current grid inpt rgn vector
  % Iterate over all blocks in current grid
  for k = 1:Nb,
    ab = alpha(inpntcx(g, k, INPT)); % Dict coefs for current block    
    ub = INPT.Phi{g}{k}*ab; % Current block reconstruction
    % Set current block part of u1 from R applied to current blk recon
    u1(inpntrx(g, k, INPT)) =  inpntR(ub, g, k, INPT);
    % Add contrib from current block to sum of QT applied to blk recons
    qtub = qtub + inpntQT(ub, g, k, INPT);
  end
  QTUB{g} = qtub; % Set QT of blk recons sum for current grid
end

% Apply G matrix to u2
for g = 2:Ng,
  u2(inpntgx(g-1, INPT)) = INPT.Grm{g-1}.*QTUB{g} - INPT.Grm{g}.*QTUB{g-1};
end

% Construct composite u from concatenation of u1 and u2
u = [sqrt(INPT.gamma0/length(u1))*u1; sqrt(INPT.gamma1/length(u2))*u2];

return


function alpha = inpntAT(u, INPT)

% inpntAT -- Apply inpainting operator A^T to compound vector
%
% Usage:
%       alpha = inpntAT(u, INPT)
%
% Input:
%       u            Compound vector resulting from operator A
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       alpha        Compound coefficient vector
  
  
Ng = size(INPT.Phi,1); % Number of grids
u1 = u(1:INPT.Nv1); % Known pixel mismatch vector
u2 = u((INPT.Nv1+1):end); % Grid overlap mismatch vector

% Initialise cell array of GT applied to each grid pair part of u2
GTU1 = cell(Ng,1);

% Compute GT applied to each grid pair in u2
GTU1{1} = -INPT.Grm{1}.*u2(inpntgx(1, INPT));
for g = 2:(Ng-1),
  GTU1{g} = INPT.Grm{g}.*(u2(inpntgx(g-1, INPT)) - u2(inpntgx(g, INPT)));
end
GTU1{Ng} = INPT.Grm{Ng}.*u2(inpntgx(Ng-1, INPT));

alpha = zeros(INPT.Nv0,1); % Initialise concatenated coef vector
% Iterate over all grids
for g = 1:Ng,
  Nb = size(INPT.Crd{g},1); % Number of blocks in current grid
  % Iterate over all blocks in current grid
  for k = 1:Nb,
    % Apply R^T to current block in u1
    rtu1b = inpntRT(u1(inpntrx(g, k, INPT)), g, k, INPT);
    % Apply Q to current block in GTU1
    qgtu2b = inpntQ(GTU1{g}, g, k, INPT);
    % Sum known pixel mismatch and grid overlap mismatch
    % contributions for current block
    ub = sqrt(INPT.gamma0/length(u1))*rtu1b + ...
	 sqrt(INPT.gamma1/length(u2))* qgtu2b;
    % Compute coefficients for current block by applying Phi^T to ub
    alpha(inpntcx(g, k, INPT)) = (INPT.Phi{g}{k})'*ub;
  end
end
    
return


function v = inpntQ(u, k, l, INPT)

% inpntQ -- Apply inpainting operator Q to inpainting region vector 
%
% Usage:
%       v = inpntQ(u, k, l, INPT)
%
% Input:
%       u            Inpainting region vector
%       k            Grid index
%       l            Index of block in grid k
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       v            Block k,l, non-zero where intersecting with
%                    inpainting region 

xd = INPT.Map{k}{l}(:,2);
xr = INPT.Map{k}{l}(:,1);
v = zeros(size(INPT.Phi{k}{l},1),1);
v(xr) = u(xd);

return


function v = inpntQT(u, k, l, INPT)

% inpntQ -- Apply inpainting operator Q to image block 
%
% Usage:
%       v = inpntQT(u, k, l, INPT)
%
% Input:
%       v            Image block k,l
%       k            Grid index
%       l            Index of block in grid k
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       v            Inpainting region vector contribution from
%                    block k,l
    
xd = INPT.Map{k}{l}(:,1);
xr = INPT.Map{k}{l}(:,2);
v = zeros(INPT.Nvg,1);
v(xr) = u(xd);

return


function v = inpntR(u, k, l, INPT)

% inpntR -- Apply inpainting operator R to image block k,l
%
% Usage:
%       v = inpntR(u, k, l, INPT)
%
% Input:
%       u            Image block k,l
%       k            Grid index
%       l            Index of block in grid k
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       v            Known part of block k,l

rg = INPT.Rgn{k}{l};
v = u(rg(:) > 0);

return


function v = inpntRT(u, k, l, INPT)

% inpntRT -- Apply inpainting operator R^T to known part of image block k,l
%
% Usage:
%       v = inpntRT(u, k, l, INPT)
%
% Input:
%       u            Known part of block k,l
%       k            Grid index
%       l            Index of block in grid k
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       v            Image block k,l


rg = INPT.Rgn{k}{l};
v = zeros(size(INPT.Phi{k}{l},1),1);
v(rg(:) > 0) = u;

return


function v = inpntRB(u, INPT)

% inpntRB -- Apply product of inpainting operators R and B to an image
%
% Usage:
%       v = inpntRB(u, INPT)
%
% Input:
%       u            Input image
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       u            Compound vector resulting from operators R and B

  
v = zeros(INPT.Nv1,1);
Ng = size(INPT.Phi,1);
for g = 1:Ng,
  Nb = size(INPT.Crd{g},1);
  for k = 1:Nb,
    rgn = INPT.Rgn{g}{k};
    r1 = INPT.Crd{g}(k,1);
    c1 = INPT.Crd{g}(k,2);
    r2 = INPT.Crd{g}(k,3);
    c2 = INPT.Crd{g}(k,4);
    ub = u(r1:r2,c1:c2,:); 
    vb = ub(rgn > 0);    
    v(INPT.Ix1{g}(k,1):INPT.Ix1{g}(k,2)) = vb;
  end
end

return


function idx = inpntrx(k, l, INPT)

% inpntrx -- Find start index of block k,l in concatenated
%            known-region vector 
%
% Usage:
%       idx = inpntrx(k, l, INPT)
%
% Input:
%       k            Grid index
%       l            Index of block in grid k
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       idx          Index in concatenated known-region vector

  
idx = INPT.Ix1{k}(l,1):INPT.Ix1{k}(l,2);

return


function idx = inpntcx(k, l, INPT)
  
% inpntcx -- Find start index of block k,l in concatenated coefficient vector
%
% Usage:
%       idx = inpntcx(k, l, INPT)
%
% Input:
%       k            Grid index
%       l            Index of block in grid k
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       idx          Index in concatenated coefficient vector
 

idx = INPT.Ix0{k}(l,1):INPT.Ix0{k}(l,2);

return


function idx = inpntgx(k, INPT)

% inpntgx -- Find start index of entry k in inpainting region grid
%            difference vector 
%
% Usage:
%       idx = inpntgx(k, INPT)
%
% Input:
%       k            Grid difference entry index
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       idx          Index in inpainting region difference vector 
 

idx = ((k-1)*INPT.Nvg+1):(k*INPT.Nvg);

return


function [img, bn, bv, grd, alp] = inpntimg(alpha, INPT)

% inpntimg -- Construct final inpainted image from a compound coefficient
%             vector
%
% Usage:
%       [img, bnm, grd, alp] = inpntimg(alpha, INPT)
%
% Input:
%       alpha        Compound coefficient vector
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       img          Inpainted image
%       bnm          Image providing count of number of grids
%                    contributing to each pixel
%       grd          Cell array of individual grid reconstructions
%       alp          Cell array of dictionary coefficients on the grid

Ng = size(INPT.Phi,1); % Number of grids
bn = zeros(size(INPT.Gms{1},1),size(INPT.Gms{1},2));
bmn = zeros(size(INPT.Gms{1}));
bm2 = zeros(size(INPT.Gms{1}));
grd = cell(Ng,2);
alp = cell(Ng,1);

% Iterate over all grids
for g = 1:Ng,
  Nb = size(INPT.Crd{g},1); % Number of blocks in current grid
  grd{g,1} = zeros(size(INPT.Gms{1}));
  grd{g,2} = zeros(size(INPT.Gms{1}));
  alp{g} = cell(Nb,1);
  % Iterate over all blocks in current grid
  for k = 1:Nb,

    av = alpha(inpntcx(g, k, INPT));
    alp{g}{k} = av;  
    %av = alpha{g}{k};
    ub = INPT.Phi{g}{k}*av;
    r1 = INPT.Crd{g}(k,1);
    c1 = INPT.Crd{g}(k,2);
    r2 = INPT.Crd{g}(k,3);
    c2 = INPT.Crd{g}(k,4);
    ubb = reshape(ub,r2-r1+1,c2-c1+1,INPT.Nih);
    if INPT.ClipGrids,
      ubb(ubb < INPT.imgrng(1)) = INPT.imgrng(1);
      ubb(ubb > INPT.imgrng(2)) = INPT.imgrng(2);
    end
    grd{g,1}(r1:r2,c1:c2,:) = ubb;
    grd{g,2}(r1:r2,c1:c2,:) = 1;
    
    bn(r1:r2,c1:c2) = bn(r1:r2,c1:c2) + 1;
    dlt = ubb - bmn(r1:r2,c1:c2,:);
    bmn(r1:r2,c1:c2,:) = bmn(r1:r2,c1:c2,:) + dlt ./ ...
        repband(bn(r1:r2,c1:c2,:),INPT.Nih);
    bm2(r1:r2,c1:c2,:) = bm2(r1:r2,c1:c2,:) + dlt .*(ubb - bmn(r1:r2,c1:c2,:));

  end

end

bv = bm2 ./ repband(bn, INPT.Nih);
img = bmn;

return
