function [alpha, optinf] = irlsbpdn(DC, s, lambda, opt, alpha0)

% irlsbpdn -- Sparse representation via IRLS method (for details
%             see P. Rodríguez and B. Wohlberg, "An Efficient
%             Algorithm for Sparse Representations with l^p Data
%             Fidelity Term", Proc. 4th IEEE Andean Technical
%             Conference (ANDESCON), (Cusco, Perú), Oct 2008).
%
% Usage:
%       [alpha, optinf] = irlsbpdn(DC, s, lambda, opt, alpha0)
%
% Input:
%       DC          Dictionary matrix or cell array of functions {Phi,PhiT}
%       s           Image to decompose
%       lambda      Regularization parameter
%       opt         Algorithm parameters structure  
%       alpha0      Initial solution
%
% Output:
%       alpha       Dictionary coefficients
%       optinf      Details of optimisation

if nargin < 5,
  alpha0 = [];
end
if nargin < 4,
  opt = [];
end
opt = defaultopts(opt);
smflg = opt.SubstMethod;

if iscell(DC),
  Phi = DC{1};
  PhiT = DC{2};
else
  Phi = @(x) DC*x;
  PhiT = @(x) DC'*x;
end

if opt.Verbose && opt.MaxMainIter > 0,
  disp('Itn  Fnc       DFid      Reg       WFnc      WDFid     WReg      RelRes    CGIt  Flg');  
  disp('------------------------------------------------------------------------------------');
end

tstart = tic;

p = opt.DatFidNorm;
q = opt.RegTrmNorm;
epsd = opt.DatFidWThresh;
epsr = opt.RegTrmWThresh;
cgt = opt.CGTol;

if isempty(alpha0),
  Wd = ones(size(Phi,1),1); % Wd is a scalar if Phi is a fun. ref.
  Wr = ones(size(Phi,2),1); % Wr is a scalar if Phi is a fun. ref.
  J0 = [];
else
  if p == 2,
    Wd = 1;
  else
    Wd = (2/p)*ft(Phi(alpha0)-s, p, epsd);
  end
  Wr = (2/q)*ft(alpha0, q, epsr);
  if opt.ZRCFraction > 0,
    a0i = abs(alpha0) > rnkval(alpha0, 1-opt.ZRCFraction);
    Wr(a0i) = 0;
  end
  J0 = sum((Phi(alpha0) - s).^p)/p + lambda*sum(abs(alpha0).^q)/q;
end

optinf = [];
optinf.itstat = [];
optinf.opt = opt;
ffc = Inf;
k = 1;

if smflg,
  if isempty(alpha0),
    beta = [];
  else
    beta = (Wr.^(1/2)).*alpha0;
  end
else
  alpha = alpha0;
end

while k <= opt.MaxMainIter & ffc > opt.FracStopTol,

  if smflg,
    A = @(x) (Wr.^(-1/2)).*PhiT(Wd.*Phi((Wr.^(-1/2)).*x)) + lambda*x;
    WPs = (Wr.^(-1/2)).*PhiT(Wd.*s);
    if ~isempty(beta),
      cgt = relres(A, WPs, beta) / opt.CGTolScale;
    end
    [beta, flg, rlr, pit] = pcg(A, WPs, cgt, opt.MaxCGIter, [], [], beta);
    alpha = (Wr.^(-1/2)).*beta;
  else
    A = @(x) PhiT(Wd.*Phi(x)) + lambda*Wr.*x;
    WPs = PhiT(Wd.*s);
    if ~isempty(alpha),
      cgt = relres(A, WPs, alpha) / opt.CGTolScale;
    end
    [alpha, flg, rlr, pit] = pcg(A, WPs, cgt, opt.MaxCGIter, [], [], alpha);
  end

  if ~isempty(opt.CGTWTFactor),
    epsd = opt.CGTWTFactor * cgt;
    epsr = opt.CGTWTFactor * cgt;
  end

  if p == 2,
    Wd = 1;
  else
    Wd = (2/p)*ft(Phi(alpha)-s, p, epsd);
  end
  Wr = (2/q)*ft(alpha, q, epsr);
  if opt.ZRCFraction > 0,
    a0i = abs(alpha) > rnkval(alpha, 1-opt.ZRCFraction);
    Wr(a0i) = 0;
  end
  
  Jdf = sum(abs(Phi(alpha) - s).^p)/p;
  Jrg = sum(abs(alpha).^q)/q;
  J1 = Jdf + lambda*Jrg;

  Jdfw = sum(((Wd.^(1/2)).*(Phi(alpha) - s)).^2)/2;
  Jrgw = sum(((Wr.^(1/2)).*alpha).^2)/2;
  J1w = Jdfw + lambda*Jrgw;

  if opt.Verbose,
    disp(sprintf('%3d %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %5d %3d', ...
                 k, J1, Jdf, Jrg, J1w, Jdfw, Jrgw,rlr,pit,flg));
  end
  optinf.itstat = [optinf.itstat; [k J1 Jdf Jrg J1w Jdfw Jrgw rlr pit flg]];

  if ~isempty(J0),
    ffc = abs((J0-J1)/J1);
  end
  % Should test for J1 > J0 here, break with warning return val if true
  J0 = J1;
  k = k + 1;

end

optinf.runtime = toc(tstart);

if opt.Verbose && opt.MaxMainIter > 0,
  
end

opt.CGTol = cgt;
opt.CGTWTFactor = epsr;
optinf.opt = opt;

if opt.Verbose && opt.MaxMainIter > 0,
  disp('------------------------------------------------------------------------------------');
end

return


function opt = defaultopts(opt)

  if ~isfield(opt,'Verbose'),
    opt.Verbose = 0;
  end
  if ~isfield(opt,'SubstMethod'),
    opt.SubstMethod = 0;
  end
  if ~isfield(opt,'DatFidNorm'),
    opt.DatFidNorm = 2;
  end
  if ~isfield(opt,'RegTrmNorm'),
    opt.RegTrmNorm = 1;
  end
  if ~isfield(opt,'DatFidWThresh'),
    opt.DatFidWThresh = 1e-5;
  end
  if ~isfield(opt,'RegTrmWThresh'),
    opt.RegTrmWThresh = 1e-5;
  end
  if ~isfield(opt,'ZRCFraction'), % Zero Regularisation Cost Fraction
    opt.ZRCFraction = 0;
  end
  if ~isfield(opt,'MaxMainIter'),
    opt.MaxMainIter = 20;
  end
  if ~isfield(opt,'FracStopTol'),
    opt.FracStopTol = 1e-3;
  end
  if ~isfield(opt,'MaxCGIter'),
    opt.MaxCGIter = 100000;
  end
  if ~isfield(opt,'CGTol'),
    opt.CGTol = 1e-2;
  end
  if ~isfield(opt,'CGTolScale'),
    opt.CGTolScale = 5;
  end
  if ~isfield(opt,'CGTWTFactor'),
    opt.CGTWTFactor = 1;
  end

return


function y = ft(x, p, thrs)

  absx = abs(x);
  absxp = absx.^(p-2);
  absxp(isnan(absxp)) = 0;
  absxp(isinf(absxp)) = 0;
  y0 = (absx > thrs).*absxp;
  y1 = (absx <= thrs).*(ones(size(absx))*thrs^(p-2));
  y = y0 + y1;

return


% Compute relative residual as used by pcg function
function y = relres(A, b, x)
  y = norm(A(x) - b)/norm(b);
return
